#include <stdio.h>
#include <stdlib.h>
#include <linux/bpf.h>
#include <bpf/bpf.h>
#include <bpf/libbpf.h>
#include <linux/perf_event.h>
#include <argp.h>
#include <sys/ioctl.h>
#include <sys/syscall.h>
#include <unistd.h>
#include <string.h>
#include <utmp.h>
#include <signal.h>
#include <time.h>
#include <pwd.h>
#include <unistd.h>
#include <fcntl.h>
// #include "skel.ebpf.h"
#include "common.h"

static int fd_map_execve;
static int fd_map_args;
static struct arguments arguments;
static FILE *f;
static int running = 1;
static char hostname[127];

struct arguments {
    char *filename;
    int to_output;
};


/*
 * Functions for arguments
 * https://www.gnu.org/software/libc/manual/html_node/Argp-Example-3.html
 */

static char doc[] = "SSH Trace usage:";
static char args_doc[] = "ARG1 ARG2";

static error_t parse_opts(int key, char *arg, struct argp_state *state){
    struct arguments *arguments = state->input;
    switch(key){
        case 'f':
            arguments->filename = arg;
            arguments->to_output = 0;
            break;
        case 'o':
            arguments->to_output = 1;
            break;
        case ARGP_KEY_ARG:
            break;
        case ARGP_KEY_END:
            break;
        default:
            return ARGP_ERR_UNKNOWN;

    }
    return 0;
}

struct arguments parse_args(int argc, char *argv[]){
    static const struct argp_option opts[] = {
        {"filename", 'f', "FILENAME", 0, "Save result to logs"},
        {"to-output", 'o', NULL, 0, "Print to output"},
        {NULL, 'h', NULL, OPTION_HIDDEN, "help"},
        {},
    };
    struct arguments arguments;
    arguments.filename = NULL;
    arguments.to_output = 1;
    static struct argp argp = {opts, parse_opts, args_doc, doc};

    argp_parse(&argp, argc, argv, 0, 0, &arguments);
    return arguments;
}

/* End functions arguments */

static void signalHandler(int signum){
    running = 0;
}

/*
 * This function get the username from the uid
 */
static void get_username(uid_t uid, char *username){
    struct passwd *p;
    p = getpwuid(uid);
    if (p == NULL){
        printf("Failed to get the username from the UID\n");
        strncpy(username, "Unknown", 8);
        return;
    }
    strncpy(username, p->pw_name, 64);
}

/*
 * With the pid, walk through the process tree and to find the sshd proc
 */
static pid_t walk_process_tree(const pid_t pid){
	char filename[64];
	FILE *fd;
	int ppid;
	char proc_name[128];

	snprintf(filename, 64, "/proc/%d/stat", pid);

	if ((fd = fopen(filename, "r")) == NULL){
		printf("Failed to open the file\n");
		return -1;
	}
	
	// 1866044 (sshd) S 1 1866044
	fscanf(fd, "%*d %s %*c %d", proc_name, &ppid);

	if (strncmp(proc_name, "(sshd)", 128) == 0){
		fclose(fd);
		return ppid;
	}

	// If parent processes
	if (ppid == 0 || ppid == 1){
		fclose(fd);
		return 0;
	}

	// printf("PID: %d; PPID: %d; Proc name: %s\n", pid, ppid, proc_name);

	fclose(fd);
	return walk_process_tree(ppid);
}

/*
 * We can find the user attached to the pid
 */
static struct utmp find_user_from_pid(pid_t pid){
    int fd;
    size_t len;
    struct utmp u = {0};

    if ((fd = open("/var/run/utmp", O_RDONLY)) < 0){
        printf("Failed to open the file\n");
        return u;
    }

    lseek(fd, 0, SEEK_SET);
    while ((len = read(fd, &u, sizeof(struct utmp))) > 0){
        if (u.ut_type == USER_PROCESS && u.ut_pid == pid)
            break;
        memset(&u, 0, sizeof(struct utmp));
    }
    close(fd);
    return u;
}
/*
 * Try to find the sshd forked in the map
 */
static struct execve find_sshd_daemon(pid_t pid){
    struct execve s_execve = {0};
    // Get the execve struct
    int err = bpf_map_lookup_elem(fd_map_execve, &pid, &s_execve);
    if (err != 0)
        return s_execve;
      
    if(strncmp(s_execve.filename, "sshd", 4) == 0)
        return s_execve;
    else
        return find_sshd_daemon(s_execve.ppid);
}
/*
 * This function get the localtime into the rsyslog format
 */
static int syslog_time(time_t ts, char t[32], size_t l){
    const char format[] = "%b %d %T";
    struct tm *lt = localtime(&ts);
    if(strftime(t, l, format, lt) == 0)
        return -1;
    return 0;
}
int handle_event(void *ctx, void *data, size_t data_sz){
    struct event *s_event = (struct event *)data;
    struct execve s_execveParent = {0};
    struct utmp u = {0};
    time_t ts = time(NULL);
    char username[64];

    // Fid the sshd process parent
    s_execveParent = find_sshd_daemon(s_event->ppid);
    if (s_execveParent.pid == 0){
        //printf("SSH daemon not exist\t%s\t%d\n", s_execve.filename, s_execve.pid);
        s_execveParent.ppid = walk_process_tree(s_event->pid); /* Walk through process tree to find the sshd daemon */
        if (s_execveParent.ppid <= 0) // We didn't find it 
            return 0;
    }
   
    // Find the user related
    u = find_user_from_pid(s_execveParent.ppid);
    if (u.ut_type == 0)
        return 0;

    // Get the username
    get_username(s_event->uid, username);

    if (arguments.to_output == 1){
        char t[32];
        syslog_time(ts, t, sizeof(t));
        printf("%-20s%-10d%s@%-25s", t, s_execveParent.ppid, u.ut_user, u.ut_host);
        printf("%-10s%-10d", username, s_event->pid);
        printf("%-20s", s_event->filename);
        printf("%s ", s_event->comm);
        // Arguments
        if (s_event->argc > 0){
           for(int i = 0; i < s_event->tlen; i++){
               char c = s_event->args[i];
               if (c == '\0')
                   printf(" ");
               else
                   printf("%c", c);
           }
        }
        printf("\n");
    }

    if (arguments.filename != NULL && f != NULL){
        /*
         * Rsyslog format
         * <time> <hostname> <procname>: <info> <data>
         */
        char t[32];
        char host[294];
        char ppid[16];
        char command[ARGS_LEN * ARGS_CNT];
        char pid[16];
        char hname[128];
        if (syslog_time(ts, t, sizeof(t)) == 0)
            fwrite(t, strlen(t), 1, f);

        snprintf(hname, 128, " %s", hostname);
        fwrite(hname, strlen(hname), 1, f);
        fwrite(" ssh-trace: ", 12, 1, f);
        fwrite("<info> ", 7, 1, f);

        snprintf(host, 294, "host=%s@%s;", u.ut_user, u.ut_host);
        fwrite(host, strlen(host), 1, f);

        snprintf(ppid, 16, "ppid=%d;", s_execveParent.ppid);
        fwrite(ppid, strlen(ppid), 1, f);

        snprintf(command, ARGS_LEN * ARGS_CNT, "pathname=%s ", s_event->filename);
        if (s_event->argc > 0){
            char output[ARGS_TLEN];
            int i;
            for(i = 0; i < s_event->tlen; i++){
                char c = s_event->args[i];
                if (c == '\0')
                    output[i] = ' ';
                else
                    output[i] = c;
           }
           output[i] = '\0';
           strncat(command, output, strlen(output));
        } 
        command[strlen(command) - 1] = ';';
        fwrite(command, strlen(command), 1, f);

        snprintf(pid, 16, "pid=%d\n", s_event->pid);
        fwrite(pid, strlen(pid), 1, f);
    }

    return 0;
}
int main(int argc, char *argv[]){
    const char *fileObj = "src/ssh-trace.ebpf.o";
    struct bpf_object *obj;
    struct bpf_program *programProcessFork;
    struct bpf_program *programExecve;
    struct bpf_program *programExitExecve;
    struct ring_buffer *rb;
    // struct sshtrace *skel;
    int err;
    int fd_map_data;

    arguments = parse_args(argc, argv); // Parsing arguments

    signal(SIGINT, signalHandler);

    /* Generate our skel for the ring buffer */
    /*skel = sshtrace__open();

    if (!skel){
        printf("Failed to init skel\n");
        return -1;
    }*/

    /* Open and load our eBPF object */
    obj = bpf_object__open_file(fileObj, NULL);
    if (!obj){
        printf("Failed to open the file\n");
        return -1;
    }

    err = bpf_object__load(obj);
    if (err){
        printf("Failed to load object\n");
        return -1;
    }


    /* Retrieving fd of maps */
    fd_map_data = bpf_object__find_map_fd_by_name(obj, "data");
    if (!fd_map_data){
        printf("Failed to find the fd map data\n");
        bpf_object__close(obj);
        return -1;
    }

    fd_map_execve = bpf_object__find_map_fd_by_name(obj, "execve_maps");
    if (!fd_map_execve){
        printf("Failed to find the fd map execve\n");
        bpf_object__close(obj);
        return -1;
    }
    fd_map_args = bpf_object__find_map_fd_by_name(obj, "execve_args");
    if (!fd_map_args){
        printf("Failed to find the fd map args\n");
        bpf_object__close(obj);
        return -1;
    }

    /* Retrieving our programs */
    programExecve = bpf_object__find_program_by_name(obj, "syscall_execve");
    programExitExecve = bpf_object__find_program_by_name(obj, "syscall_exit_execve");
    programProcessFork = bpf_object__find_program_by_name(obj, "syscall_process_fork");

    if (!programExitExecve || !programExecve || !programProcessFork){
        printf("Failed to find program\n");
        bpf_object__close(obj);
        return -1;
    }

    bpf_program__attach(programProcessFork);
    bpf_program__attach(programExecve);
    bpf_program__attach(programExitExecve);

    /* Start the ringbuffer */
    rb = ring_buffer__new(fd_map_data, handle_event, NULL, NULL);
    if (!rb){
        printf("Failed to create the ringbuf\n");
        bpf_object__close(obj);
        return -1;
    }

    if (arguments.to_output == 1)
        printf("%-20s%-10s%-30s%-10s%-10s%-20s%s\n", "Datetime", "PID sshd", "Username@IP Address", "Username", "Pid", "Pathname", "Command");

    if (arguments.filename != NULL){
        f = fopen(arguments.filename, "a");
        if (f == NULL){
            printf("Failed to create the file %s\n", arguments.filename);
            return 0;
        }
    }

    /* Get the hostname */
    if (gethostname(hostname, 127) == -1){
        printf("Failed to get the hostname\n");
        strncpy(hostname, "ubuntu", 7);
    }

    while(running){
        err = ring_buffer__poll(rb, 100 /* timeout, ms */);
        if (err == -EINTR){
            printf("Failed to get the ringbuf\n");
            running = 0;
            break;
         }
    }

    if (f != NULL)
        fclose(f);
    ring_buffer__free(rb);
    bpf_object__close(obj);

    return 0;
}
