#define BPF_NO_GLOBAL_DATA
#define __TARGET_ARCH_x86
#include "vmlinux.h"
#include <bpf/bpf_helpers.h>
#include <bpf/bpf_tracing.h>
#include <bpf/bpf_core_read.h>
#include <bpf/bpf_endian.h>
#include <string.h>
#include "common.h"

/*
 * Helper:
 * Issue: invalid indirect read from stack R2 off
 *   Fix: check if all variables is initialised
 * Issue: R1 invalid mem access 'inv'
 *   Fix: the value can be NULL
 */

struct {
    __uint(type, BPF_MAP_TYPE_RINGBUF);
    __uint(max_entries, 256 * 1024 /* 256kb */);
} data SEC(".maps");

struct {
    __uint(type, BPF_MAP_TYPE_HASH);
    __uint(max_entries, 32768); // pid_max -> https://linux.die.net/man/5/proc
    __type(key, pid_t);
    __type(value, struct execve);
} execve_maps SEC(".maps");

struct ctx_execve {
   __u16 common_type;          // Unsigned short
   __u8 common_flags;          // Unsigned char
   __u8 common_preempt_count;  // Unsigned char
   __s32 common_pid;           // int
   __s32 syscall_nr;           // int
   const __u8 *filename;       // unsigned char
   const __u8 *const *argv;
   const __u8 *const *envp;
};

struct ctx_fork {
   __u16 common_type;          // Unsigned short
   __u8 common_flags;          // Unsigned char
   __u8 common_preempt_count;  // Unsigned char
    __s32 common_pid;
    char parent_comm[16];
    pid_t parent_pid;
    char child_comm[16];
    pid_t child_pid;
};

/*
 * Made my own strncmp, because the function bpf_strncmp not available
 */
static int b_strncmp(const char *s1, __s32 len, const char *s2){
    __s32 c = 0;
    __s32 i;
    for (i = 0; i < len; i++){
    //while (c != '\0' || i != len){
        //c = s1[i];
        if (s1[i] == '\0')
            break;

        if (s1[i] == s2[i])
            c++;
    }
    return (c == i) ? 0: 1;
}
/*
 * Need to detect the sshd forked, because, when a new user is connected
 * we cannot detect the execve sshd, but only the fork
 * Execve of the sshd is executed when the user is connected
 * But, if the user is authenticated, the process is forked
 */
SEC("tp/sched/sched_process_fork")
int syscall_process_fork(struct ctx_fork *ctx){
	pid_t ppid;
    char child[16];
	pid_t pid = bpf_get_current_pid_tgid() >> 32;
	struct task_struct *t = (struct task_struct*)bpf_get_current_task();
	struct task_struct *real_parent;
	bpf_probe_read(&real_parent, sizeof(struct task_struct *), &t->real_parent);

	bpf_probe_read(&ppid, sizeof(pid_t), &real_parent->pid);
    bpf_probe_read(&child, 16, &ctx->child_comm);

    if (b_strncmp(child, 16, "sshd") == 0){
        struct execve s_execve = {0};
        s_execve.pid = pid;
        s_execve.ppid = ppid;
        bpf_probe_read(&s_execve.filename, sizeof(s_execve.filename), ctx->child_comm);
        // bpf_printk("Fork: %d %d", s_execve.pid, s_execve.ppid);
    
        if (bpf_map_update_elem(&execve_maps, &pid, &s_execve, BPF_ANY) < 0){
            bpf_printk("Failed to update the map");
            return 0;
        }
    }
	return 0;
}

SEC("tp/syscalls/sys_exit_execve")
int syscall_exit_execve(void) {
    pid_t ppid;
	struct task_struct *t = (struct task_struct*)bpf_get_current_task();
	struct task_struct *real_parent;
	bpf_probe_read(&real_parent, sizeof(struct task_struct *), &t->real_parent);
	bpf_probe_read(&ppid, sizeof(pid_t), &real_parent->tgid);
    // bpf_printk("Exit execve: %d %d", pid, ppid);
    return 0;
}
static pid_t walk_process_tree(struct task_struct *parent, pid_t pid){
    struct task_struct *r;
    pid_t p;

    if(bpf_probe_read(&r, sizeof(struct task_struct *), &parent->real_parent) < 0)
        return 0;
    if(bpf_probe_read(&p, sizeof(pid_t), &r->tgid) < 0)
        return 0;
    return p; 
}
static int execve(struct ctx_execve *ctx) {
    struct event *s_event;
    __u32 tlen = 0;
    const char *argv;
    // kuid_t kuid; /* https://elixir.bootlin.com/linux/v6.11.8/source/include/linux/uidgid_types.h#L9 */

    s_event = bpf_ringbuf_reserve(&data, sizeof(*s_event), 0);
    if (!s_event)
        return 0;

    s_event->pid = bpf_get_current_pid_tgid();
    s_event->ppid = 0;
    s_event->uid = bpf_get_current_uid_gid();

    bpf_probe_read_user_str(&s_event->filename, sizeof(s_event->filename), ctx->filename);

    // Arguments
    s_event->argc = 0;
    s_event->tlen = 0;

    // Take the command executed
    long err = bpf_probe_read_user_str(&argv, sizeof(argv), &ctx->argv[0]);
    if (err < 0){
        bpf_printk("Failed to get the command");
    }
    else{
        __s32 len = bpf_probe_read_user_str(&s_event->comm, COMM_LEN, argv);
        if (len > COMM_LEN)
            bpf_printk("Cannot get the command");
    }
    // Start at one, because 0 is the command
    for (__s32 i = 1; i < ARGS_CNT; i++){
        err = bpf_probe_read_user_str(&argv, sizeof(argv), &ctx->argv[i]);
        
        if (err < 0){
            //bpf_ringbuf_discard(s_event, 0);
            bpf_printk("Failed to read args");
            break;
        }

        /*
         * To avoid error:
         *  Unbounded memory access, make sure to bounds check any such access
         */
        if (tlen > (ARGS_TLEN - ARGS_LEN)){
            //bpf_ringbuf_discard(s_event, 0);
            bpf_printk("Max arguments exceeded: %d %d", i, tlen);
            break;
        }

        __u32 len = bpf_probe_read_user_str(&s_event->args[tlen], ARGS_LEN, argv);
        if (len > ARGS_LEN){
            break;
        }
        tlen += len;
        s_event->argc++;
    }
    s_event->tlen = tlen;

    // Get ppid
    struct task_struct *t = (struct task_struct*)bpf_get_current_task();
    s_event->ppid = walk_process_tree(t, s_event->pid);

    /*if (bpf_ringbuf_output(&data, &s_event, sizeof(struct event), 0) != 0){
        bpf_printk("Failed to push into the ringbuf");
        return 0;
    }*/
    bpf_ringbuf_submit(s_event, 0);

    return 0;
}
SEC("tp/syscalls/sys_enter_execve")
int syscall_execve(struct ctx_execve *ctx) {
    return execve(ctx);
}
/*SEC("tp/syscalls/sys_enter_mmap")
int syscall_mmap(void){
    bpf_printk("mmap detected");
    return 0;
}*/

char LICENSE[] SEC("license") = "GPL";
